#!/usr/bin/env python

"""
Convert a bit score to an e-value.

Example usage:

$ bit-score-to-evalue.py --dbSize 168142520 --dbSequenceCount 5660 \
  --queryLength 111 --lengthAdjustment 26 --bitScore 37.3537
0.0813077725194
"""

from __future__ import print_function

import argparse

from dark.blast.score import bitScoreToEValue


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Convert a bit score to an e-value.')

    parser.add_argument(
        '--bitScore', type=float, required=True,
        help='The bit score to convert')

    parser.add_argument(
        '--dbSize', type=int, required=True,
        help='The total number of bases in the sequence database.')

    parser.add_argument(
        '--dbSequenceCount', type=int, required=True,
        help='The number of sequences in the database.')

    parser.add_argument(
        '--queryLength', type=int, required=True,
        help='The length of the query sequence.')

    parser.add_argument(
        '--lengthAdjustment', type=int, required=True,
        help='The length adjustment.')

    args = parser.parse_args()
    print(bitScoreToEValue(args.bitScore, args.dbSize, args.dbSequenceCount,
                           args.queryLength, args.lengthAdjustment))



#!/usr/bin/env python

from __future__ import print_function

from Bio import SeqIO
from collections import defaultdict
import sys
from dark.distance import levenshtein
from math import log10, ceil

# The name of the unknown adaptor.
UNKNOWN = 'UNKNOWN'


def splitFASTAByAdaptor(knownAdaptors, adaptorLen, adaptorOffset,
                        maximumDistance, outputPrefix, dryRun, verbose):
    """
    @param knownAdaptors: A C{set} of expected adaptor sequences.
    @param adaptorLen: The C{int} length of each adaptor sequence.
    @param adaptorOffset: The zero-based C{int} offset of the adaptor in
        each sequence.
    @param maximumDistance: The maximum distance an unknown adaptor will be
        mapped to in an attempt to find its nearest known adaptor.
    @param outputPrefix: A C{str} prefix that should be used in the file names
        that are written out.
    @param dryRun: A C{bool}, if C{True} only print what would be done, don't
        create any new FASTA files.
    @param verbose: A C{bool}, if C{True} output additional information about
        adaptor classes found and assigned.
    """
    adaptors = defaultdict(int)
    unknowns = 0
    classes = dict(zip(knownAdaptors, knownAdaptors))
    reads = []

    for count, seq in enumerate(SeqIO.parse(sys.stdin, 'fasta'), start=1):
        reads.append(seq)
        adaptor = str(seq.seq)[adaptorOffset:][:adaptorLen].upper()
        adaptors[adaptor] += 1

    order = sorted(adaptors, key=lambda adaptor: adaptors[adaptor],
                   reverse=True)

    for adaptor in order:
        if adaptor in knownAdaptors:
            if verbose:
                print('%s: %s. Known adaptor' % (adaptor, adaptors[adaptor]))
        else:
            distances = sorted((levenshtein(adaptor, known), known) for
                               known in knownAdaptors)
            # Treat the read as unclassifiable if it's too far from its
            # nearest neighbor or if its nearest neighbor is ambiguous.
            nearest = distances[0][0]
            if nearest > maximumDistance or (len(knownAdaptors) > 1 and
                                             nearest == distances[1][0]):
                unknowns += 1
                classes[adaptor] = UNKNOWN
                if verbose:
                    print('%s: %s. Unknown, distances %r' % (
                        adaptor, adaptors[adaptor], [d[0] for d in distances]))
            else:
                correctedAdaptor = distances[0][1]
                classes[adaptor] = correctedAdaptor
                if verbose:
                    print('%s: %s. Assigned to class %s, at dist %d' % (
                        adaptor, adaptors[adaptor], correctedAdaptor,
                        distances[0][0]))

    readGroups = defaultdict(list)

    # Collect reads into classes.
    for read in reads:
        adaptor = str(read.seq)[adaptorOffset:][:adaptorLen].upper()
        readGroups[classes[adaptor]].append(read[adaptorOffset + adaptorLen:])

    # Calculate the number of digits in the size of the biggest read group
    # so we can nicely align the output.
    width = int(ceil(log10(max(len(group) for group in readGroups.values()))))

    # The width of the count of files we'll write, so file names have zero
    # padded numeric prefixes.
    filesWidth = int(ceil(log10(len(readGroups))))

    # Write out the FASTA files for each adaptor class (this includes the
    # unclassifiable reads if any unknown adaptors were found).
    for count, adaptor in enumerate(sorted(readGroups), start=1):
        reads = readGroups[adaptor]
        filename = '%s%0*d-%s.fasta' % (outputPrefix, filesWidth, count,
                                        adaptor)
        description = ('unrecognized adaptors' if adaptor == UNKNOWN
                       else 'adaptor %s' % adaptor)
        if dryRun:
            print('Would write %*d sequences for %s to %s' % (
                width, len(reads), description, filename))
        else:
            with open(filename, 'w') as fp:
                SeqIO.write(reads, fp, 'fasta')
            print('Wrote %*d sequences for %s to %s' % (
                width, len(reads), description, filename))


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description=('From a set of known adaptors and a FASTA on stdin, '
                     'write FASTA files split by adaptor'))

    parser.add_argument(
        'adaptors', nargs='+', metavar='adaptor',
        help='the set of adaptors that were used in sequencing')

    parser.add_argument(
        '--adaptor-offset', metavar='N',
        type=int, default=0, dest='adaptorOffset',
        help='the offset of the adaptor in each sequence')

    parser.add_argument(
        '--maximum-distance', metavar='dist',
        type=int, default=2, dest='maximumDistance',
        help='The maximum distance an unknown adaptor will be '
             'mapped to in an attempt to find its nearest known adaptor.')

    parser.add_argument(
        '--outputPrefix', metavar='prefix', default='MID-',
        help='the prefix to use for newly created FASTA files')

    parser.add_argument(
        '--dry-run', type=bool, default=False, dest='dryRun',
        help='If True, do not write new FASTA files, just show what would be '
        'done')

    parser.add_argument(
        '--verbose', type=bool, default=False,
        help='If True, print information about adaptor classes')

    args = parser.parse_args()
    adaptorLen = len(args.adaptors[0])
    adaptors = set(args.adaptors)

    # Check all adaptors are the same length.
    if any((len(adaptor) != adaptorLen) for adaptor in adaptors):
        print('All adaptors must be the same length.', file=sys.stderr)
        sys.exit(1)

    splitFASTAByAdaptor(adaptors, adaptorLen, args.adaptorOffset,
                        args.maximumDistance, args.outputPrefix, args.dryRun,
                        args.verbose)



import re

from dark.taxonomy import LineageFetcher
from dark.filter import TitleFilter
from dark.score import HigherIsBetterScore


def bestAlignment(readAlignments):
    """
    Find the best alignment for a read. This is the one whose first HSP
    has the best score.

    Note that the comparison of HSP score values is taken care of by
    the HSP class. This works whether higher or lower scores are
    considered better.

    @param readAlignments: a C{ReadAlignments} instance.
    @return: The alignment with the best first HSP score.

    """
    return max(readAlignments, key=lambda alignment: alignment.hsps[0])


class Alignment(object):
    """
    Hold information about a read alignment.

    @param subjectLength: The C{int} length of the sequence a read matched
        against.
    @param subjectTitle: The C{str} title of the sequence a read matched
        against.
    """

    def __init__(self, subjectLength, subjectTitle):
        self.subjectLength = subjectLength
        self.subjectTitle = subjectTitle
        self.hsps = []

    def addHsp(self, hsp):
        """
        Add an HSP to the list of HSPs for this alignment.

        @param hsp: A L{dark.hsp} (or subclass) instance.
        """
        self.hsps.append(hsp)


class ReadAlignments(list):
    """
    Holds information about the alignments for a read.

    @param read: A C{Read} instance.
    @param alignments: A C{list} of L{dark.alignment.Alignment} instances or
        C{None} if the read has no alignments.
    """
    def __init__(self, read, alignments=None):
        list.__init__(self)
        self.read = read
        if alignments:
            self.extend(alignments)


class ReadsAlignmentsParams(object):
    """
    Holds information about how a ReadsAlignments instance was created.

    @param application: The C{str} name of the application that created
        the ReadsAlignments data.
    @param applicationParams: A C{dict} holding options and their values
        given to the application that created the ReadsAlignments data.
    @param subjectIsNucleotides: A C{bool} that indicates whether subject
        sequences (that are aligned against) are at the nucleotide level.
        If C{False}, the subject is assumed to be amino acids.
    @param scoreTitle: A C{str} to describe a score. E.g, "Bit value"
        or "E value".
    """

    def __init__(self, application, applicationParams=None,
                 subjectIsNucleotides=True, scoreTitle='Score'):
        self.application = application
        self.applicationParams = applicationParams
        self.subjectIsNucleotides = subjectIsNucleotides
        self.scoreTitle = scoreTitle


class ReadsAlignmentsFilter(object):
    """
    Provide a filter for C{ReadsAlignments} instances.

    @param limit: An C{int} limit on the number of records to read.
    @param maxAlignmentsPerRead: An C{int} limit on the number of alignments a
        read may have in order not to be filtered out. Reads with a greater
        number of alignments will be elided. Pass 0 to filter out reads that
        did not match (i.e., align to) any subjects. Use C{None} for no
        max alignments filtering.
    @param minSequenceLen: Sequences of lesser length will be elided.
    @param maxSequenceLen: Sequences of greater length will be elided.
    @param minStart: HSPs that start before this offset in the matched
        sequence should not be returned.
    @param maxStop: HSPs that end after this offset in the matched sequence
        should not be returned.
    @param oneAlignmentPerRead: If C{True}, only keep the best
        alignment for each read.
    @param maxHspsPerHit: The maximum number of HSPs to keep for each
        alignment for each read.
    @param scoreCutoff: A C{float} score. Matches with scores that are not
        better than this score will be ignored.
    @param whitelist: If not C{None}, a set of exact titles that are always
        acceptable (though the match info for a whitelist title may rule it
        out for other reasons).
    @param blacklist: If not C{None}, a set of exact titles that are never
        acceptable.
    @param titleRegex: A regex that sequence titles must match.
    @param negativeTitleRegex: A regex that sequence titles must not match.
    @param truncateTitlesAfter: A string that titles will be truncated
        beyond. If a truncated title has already been seen, that title will
        be elided.
    @param taxonomy: Either a C{str} name or an C{int} id of the taxonomic
        group on which should be filtered. eg 'Vira' will filter on
        viruses, while 11118 will filter on Coronaviridae.
    @param readIdRegex: A case-sensitive regex C{str} that read ids must
        match.
    @return: C{self}.
    """
    def __init__(self, limit=None, maxAlignmentsPerRead=None,
                 minSequenceLen=None, maxSequenceLen=None,
                 minStart=None, maxStop=None,
                 oneAlignmentPerRead=False, maxHspsPerHit=None,
                 scoreCutoff=None, whitelist=None, blacklist=None,
                 titleRegex=None, negativeTitleRegex=None,
                 truncateTitlesAfter=None, taxonomy=None, readIdRegex=None):

        self.limit = limit
        self.maxAlignmentsPerRead = maxAlignmentsPerRead
        self.minSequenceLen = minSequenceLen
        self.maxSequenceLen = maxSequenceLen
        self.minStart = minStart
        self.maxStop = maxStop
        self.oneAlignmentPerRead = oneAlignmentPerRead
        self.maxHspsPerHit = maxHspsPerHit
        self.scoreCutoff = scoreCutoff

        # If we've been asked to filter on matched sequence titles in any way,
        # build a title filter.
        if (whitelist or blacklist or titleRegex or negativeTitleRegex or
                truncateTitlesAfter):
            self.titleFilter = TitleFilter(
                whitelist=whitelist, blacklist=blacklist,
                positiveRegex=titleRegex, negativeRegex=negativeTitleRegex,
                truncateAfter=truncateTitlesAfter)
        else:
            self.titleFilter = None

        if taxonomy is not None:
            self.lineageFetcher = LineageFetcher()
        else:
            self.lineageFetcher = None
        self.taxonomy = taxonomy

        if readIdRegex is None:
            self.readIdRegex = None
        else:
            self.readIdRegex = re.compile(readIdRegex)

        self.count = 0

    def filter(self, readAlignments):
        """
        Filter a read's alignments.

        @param readAlignments: A C{ReadAlignments} instance.
        @return: A C{ReadAlignments} instance if the passed
            C{readAlignments} is not filtered out, else C{False}.
        """

        # Implementation notes:
        #
        # 1. The order in which we carry out the filtering actions can make
        #    a big difference in the result of this function. The current
        #    ordering is based on what seems reasonable - it may not be the
        #    best way to do things. E.g., if maxHspsPerHit is 1 and there
        #    is a title regex, which should we perform first?
        #
        #    We perform filtering based on alignment before that based on
        #    HSPs. That's because there's no point filtering all HSPs for
        #    an alignment that we end up throwing away anyhow.
        #
        # 2. This function could be made faster if it first looked at its
        #    arguments and dynamically created an acceptance function
        #    (taking a readAlignments as an argument). The acceptance
        #    function would run without examining the desired filtering
        #    settings on each call the way the current code does.
        #
        # 3. A better approach with readIdRegex would be to allow the
        #    passing of a regex object. Then the caller would make the
        #    regex with whatever flags they liked (e.g., case insensitive).

        #
        # Alignment-only (i.e., non-HSP based) filtering.
        #
        if self.limit is not None and self.count == self.limit:
            return False

        # Does the read have too many alignments?
        if (self.maxAlignmentsPerRead is not None and
                len(readAlignments) > self.maxAlignmentsPerRead):
            return False

        # Filter on the read id.
        if (self.readIdRegex and
                self.readIdRegex.search(readAlignments.read.id) is None):
            return False

        if self.titleFilter:
            # Remove alignments against sequences whose titles are
            # unacceptable.
            wantedAlignments = []
            for alignment in readAlignments:
                if (self.titleFilter.accept(alignment.subjectTitle) !=
                        TitleFilter.REJECT):
                    wantedAlignments.append(alignment)
            if wantedAlignments:
                readAlignments[:] = wantedAlignments
            else:
                return False

        # Only return alignments that are against sequences of the
        # desired length.
        minSequenceLen = self.minSequenceLen
        maxSequenceLen = self.maxSequenceLen
        if minSequenceLen is not None or maxSequenceLen is not None:
            wantedAlignments = []
            for alignment in readAlignments:
                length = alignment.subjectLength
                if not ((minSequenceLen is not None and
                         length < minSequenceLen) or
                        (maxSequenceLen is not None and
                         length > self.maxSequenceLen)):
                    wantedAlignments.append(alignment)
            if wantedAlignments:
                readAlignments[:] = wantedAlignments
            else:
                return False

        if self.taxonomy is not None:
            wantedAlignments = []
            for alignment in readAlignments:
                lineage = self.lineageFetcher.lineage(alignment.subjectTitle)
                if lineage:
                    for taxonomyIdAndScientificName in lineage:
                        if self.taxonomy in taxonomyIdAndScientificName:
                            wantedAlignments.append(alignment)
                else:
                    # No lineage info was found. Keep the alignment
                    # since we can't rule it out.  We could add another
                    # option to control this.
                    wantedAlignments.append(alignment)
            if wantedAlignments:
                readAlignments[:] = wantedAlignments
            else:
                return False

        if self.oneAlignmentPerRead and readAlignments:
            readAlignments[:] = [bestAlignment(readAlignments)]

        #
        # From here on we do only HSP-based filtering.
        #

        # Throw out any unwanted HSPs due to maxHspsPerHit.
        if self.maxHspsPerHit is not None:
            for alignment in readAlignments:
                hsps = alignment.hsps
                if len(hsps) > self.maxHspsPerHit:
                    alignment.hsps = hsps[:self.maxHspsPerHit]

        # Throw out HSPs whose scores are not good enough.
        if self.scoreCutoff is not None:
            wantedAlignments = []
            for alignment in readAlignments:
                hsps = alignment.hsps
                wantedHsps = []
                for hsp in hsps:
                    if hsp.betterThan(self.scoreCutoff):
                        wantedHsps.append(hsp)
                if wantedHsps:
                    alignment.hsps = wantedHsps
                    wantedAlignments.append(alignment)
            if wantedAlignments:
                readAlignments[:] = wantedAlignments
            else:
                return False

        # Throw out HSPs that don't match in the desired place on the
        # matched sequence.
        minStart = self.minStart
        maxStop = self.maxStop
        if minStart is not None or maxStop is not None:
            wantedAlignments = []
            for alignment in readAlignments:
                hsps = alignment.hsps
                wantedHsps = []
                for hsp in hsps:
                    if not ((minStart is not None and
                             hsp.readStartInSubject < minStart) or
                            (maxStop is not None and
                             hsp.readEndInSubject > maxStop)):
                        wantedHsps.append(hsp)
                if wantedHsps:
                    alignment.hsps = wantedHsps
                    wantedAlignments.append(alignment)
            if wantedAlignments:
                readAlignments[:] = wantedAlignments
            else:
                return False

        self.count += 1
        return readAlignments

    def close(self):
        """
        Close our lineage fetcher, if any.
        """
        if self.taxonomy:
            self.lineageFetcher.close()


class ReadsAlignments(object):
    """
    Provide for filtering for a collection of reads and their alignments.

    You probably will not use this class directly. Instead, write a
    subclass that implements iter.

    See L{dark.blast.alignments.BlastReadsAlignments} and
    L{dark.diamond.alignments.DiamondReadsAlignments} for examples.

    @param reads: A L{Reads} instance, containing the reads (query sequences)
        given to the application to create these matches.
    @param params: An instance of C{ReadsAlignmentsParams}, containing
        the details of the application that created the alignments.
    @param scoreClass: A class to hold and compare scores (see scores.py).
        Default is C{HigherIsBetterScore}, for comparing bit scores. If you
        are using e.g., BLAST e-values, pass LowerIsBetterScore instead.
    """

    def __init__(self, reads, params, scoreClass=HigherIsBetterScore):
        self.reads = reads
        self.params = params
        self.scoreClass = scoreClass
        self._filters = []

    def getSubjectSequence(self, title):
        """
        Obtain information about a sequence given its title.

        Must be implemented by a subclass, e.g., see
        L{blast.alignments.BlastReadsAlignments}.

        @param title: A C{str} sequence title from a BLAST or DIAMOND (etc.)
            match. Usually of the form
            'gi|63148399|gb|DQ011818.1| Description...'.
        @raise NotImplementedError: This method must be implemented by a
            subclass.
        """
        raise NotImplementedError('getSubjectSequence must be implemented by '
                                  'a subclass')

    def hsps(self):
        """
        Provide access to all HSPs for all alignments of all reads.

        @return: A generator that yields HSPs (or LSPs).
        """
        for readAlignments in self:
            for alignment in readAlignments:
                for hsp in alignment.hsps:
                    yield hsp

    def __iter__(self):
        """
        Iterate through all readsAlignments, yielding those that pass any
        filters that have been added.

        @return: A generator that yields readsAlignments.
        """
        for readsAlignments in self.iter():
            for filterFunc in self._filters:
                filteredReadsAlignments = filterFunc(readsAlignments)
                if filteredReadsAlignments is False:
                    break
                else:
                    readsAlignments = filteredReadsAlignments
            else:
                yield readsAlignments

    def iter(self):
        """
        Placeholder to allow subclasses to provide readsAlignments.

        These might be extracted from a file. E.g., the
        C{dark.blast.alignments.blast.BlastReadsAlignments} class (a subclass
        of C{ReadsAlignments}) overrides this method to provide readsAlignments
        from files.

        @return: An iterable of C{ReadsAlignments} instances.
        """
        return []

    def filter(self, **kwargs):
        """
        Add a filter to this C{readsAlignments}.

        @param kwargs: Keyword arguments, as accepted by
            C{ReadsAlignmentsFilter}.
        @return: C{self}
        """
        self._filters.append(ReadsAlignmentsFilter(**kwargs).filter)
        return self



from unittest import TestCase
from collections import Counter

from dark.intervals import OffsetAdjuster, ReadIntervals
from dark.hsp import HSP


class TestReadIntervals(TestCase):
    """
    Test the ReadIntervals class
    """

    EMPTY = ReadIntervals.EMPTY
    FULL = ReadIntervals.FULL

    def testEmpty(self):
        """
        When no intervals are added, walk should just return one empty
        interval that spans the entire rangoe from 0 to the sequence
        length.
        """
        ri = ReadIntervals(100)
        self.assertEqual(
            [
                (self.EMPTY, (0, 100))
            ],
            list(ri.walk()))

    def testOneIntervalExactCovering(self):
        """
        If there is a single interval that spans the whole hit exactly, just
        that one interval should be returned by walk, and it should be full.
        """
        ri = ReadIntervals(100)
        ri.add(0, 100)
        self.assertEqual(
            [
                (self.FULL, (0, 100))
            ],
            list(ri.walk()))

    def testOneIntervalCoveringAllExtendingLeft(self):
        """
        If there is a single interval that spans the whole hit, including
        going negative to the left, that one interval should be returned by
        walk, and it should be full.
        """
        ri = ReadIntervals(100)
        ri.add(-10, 100)
        self.assertEqual(
            [
                (self.FULL, (-10, 100))
            ],
            list(ri.walk()))

    def testOneIntervalCoveringAllExtendingRight(self):
        """
        If there is a single interval that spans the whole hit, including
        going beyond the hit to the right, that one interval should be returned
        by walk, and it should be full.
        """
        ri = ReadIntervals(100)
        ri.add(0, 110)
        self.assertEqual(
            [
                (self.FULL, (0, 110))
            ],
            list(ri.walk()))

    def testOneIntervalCoveringAllExtendingBoth(self):
        """
        If there is a single interval that spans the whole hit, including
        starting before zero and also going beyond the hit to the right, that
        one interval should be returned by walk, and it should be full.
        """
        ri = ReadIntervals(100)
        ri.add(-10, 110)
        self.assertEqual(
            [
                (self.FULL, (-10, 110))
            ],
            list(ri.walk()))

    def testOneIntervalStartingAtZero(self):
        """
        If there is a single interval that starts at zero but doesn't
        cover the whole hit, we should get 2 intervals back from walk,
        a full one and then an empty.
        """
        ri = ReadIntervals(100)
        ri.add(0, 50)
        self.assertEqual(
            [
                (self.FULL, (0, 50)),
                (self.EMPTY, (50, 100)),
            ],
            list(ri.walk()))

    def testOneIntervalStartingBeforeZero(self):
        """
        If there is a single interval that starts before zero but doesn't
        cover the whole hit, we should get 2 intervals back from walk,
        a full one and then an empty.
        """
        ri = ReadIntervals(100)
        ri.add(-50, 50)
        self.assertEqual(
            [
                (self.FULL, (-50, 50)),
                (self.EMPTY, (50, 100)),
            ],
            list(ri.walk()))

    def testOneIntervalEndingAtHitEnd(self):
        """
        If there is a single interval that ends at the end of the hit
        but doesn't start at zero, we should get 2 intervals back from walk,
        an empty then a full.
        """
        ri = ReadIntervals(100)
        ri.add(50, 100)
        self.assertEqual(
            [
                (self.EMPTY, (0, 50)),
                (self.FULL, (50, 100)),
            ],
            list(ri.walk()))

    def testOneIntervalEndingAfterHitEnd(self):
        """
        If there is a single interval that ends after the end of the hit
        but doesn't start at zero, we should get 2 intervals back from walk,
        an empty then a full.
        """
        ri = ReadIntervals(100)
        ri.add(50, 150)
        self.assertEqual(
            [
                (self.EMPTY, (0, 50)),
                (self.FULL, (50, 150)),
            ],
            list(ri.walk()))

    def testOneIntervalInMiddle(self):
        """
        If there is a single interval in the middle of the hit, we
        should get 3 intervals back from walk: empty, full, empty.

        """
        ri = ReadIntervals(100)
        ri.add(50, 60)
        self.assertEqual(
            [
                (self.EMPTY, (0, 50)),
                (self.FULL, (50, 60)),
                (self.EMPTY, (60, 100)),
            ],
            list(ri.walk()))

    def testTwoOverlappingIntervalsInMiddle(self):
        """
        If there are two overlapping intervals in the middle of the hit, we
        should get 3 intervals back from walk: empty, full, empty.
        """
        ri = ReadIntervals(100)
        ri.add(50, 60)
        ri.add(55, 70)
        self.assertEqual(
            [
                (self.EMPTY, (0, 50)),
                (self.FULL, (50, 70)),
                (self.EMPTY, (70, 100)),
            ],
            list(ri.walk()))

    def testThreeOverlappingIntervalsInMiddle(self):
        """
        If there are three overlapping intervals in the middle of the hit, we
        should get 3 intervals back from walk: empty, full, empty.
        """
        ri = ReadIntervals(100)
        ri.add(50, 60)
        ri.add(65, 75)
        ri.add(55, 70)
        self.assertEqual(
            [
                (self.EMPTY, (0, 50)),
                (self.FULL, (50, 75)),
                (self.EMPTY, (75, 100)),
            ],
            list(ri.walk()))

    def testPairOfTwoOverlappingIntervals(self):
        """
        If there are two sets of two overlapping intervals in the middle of
        the hit, we should get 5 intervals back from walk.
        """
        ri = ReadIntervals(100)
        # First overlapping pair, 50-70.
        ri.add(50, 60)
        ri.add(55, 70)
        # First overlapping pair, 80-95.
        ri.add(80, 90)
        ri.add(85, 95)
        self.assertEqual(
            [
                (self.EMPTY, (0, 50)),
                (self.FULL, (50, 70)),
                (self.EMPTY, (70, 80)),
                (self.FULL, (80, 95)),
                (self.EMPTY, (95, 100)),
            ],
            list(ri.walk()))

    def testOverlappingIntervalsThatCoverEverything(self):
        """
        If there are sets of overlapping intervals that cover the whole hit,
        we should get 1 full interval back from walk.
        """
        ri = ReadIntervals(100)
        ri.add(-10, 20)
        ri.add(15, 40)
        ri.add(40, 70)
        ri.add(66, 89)
        ri.add(77, 93)
        ri.add(70, 110)
        self.assertEqual(
            [
                (self.FULL, (-10, 110))
            ],
            list(ri.walk()))

    # The following tests have the same setup as above, but they test the
    # coverageCounts() method.

    def testEmptyCoverageCountsOnZeroLengthSequence(self):
        """
        When no intervals are added, coverageCounts should return an empty
        Counter.
        """
        ri = ReadIntervals(0)
        self.assertEqual({}, ri.coverageCounts())

    def testEmptyCoverageCounts(self):
        """
        When no intervals are added, coverageCounts should return an empty
        Counter.
        """
        ri = ReadIntervals(100)
        self.assertEqual({}, ri.coverageCounts())

    def testOneIntervalExactCoveringCoverageCounts(self):
        """
        If there is a single interval that spans the whole hit exactly,
        coverageCounts should return the correct result.
        """
        ri = ReadIntervals(10)
        ri.add(0, 10)
        c = Counter([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(c, ri.coverageCounts())

    def testOneIntervalCoveringAllExtendingLeftCoverageCounts(self):
        """
        If there is a single interval that spans the whole hit, including
        going negative to the left, coverageCounts should return the correct
        result.
        """
        ri = ReadIntervals(10)
        ri.add(-2, 10)
        c = Counter([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(c, ri.coverageCounts())

    def testOneIntervalCoveringAllExtendingRightCoverageCounts(self):
        """
        If there is a single interval that spans the whole hit, including
        going beyond the hit to the right, coverageCounts should return the
        correct result.
        """
        ri = ReadIntervals(10)
        ri.add(0, 12)
        c = Counter([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(c, ri.coverageCounts())

    def testOneIntervalStartingAtZeroCoverageCounts(self):
        """
        If there is a single interval that starts at zero but doesn't
        cover the whole hit, coverageCounts should return the correct result.
        """
        ri = ReadIntervals(10)
        ri.add(0, 5)
        c = Counter([0, 1, 2, 3, 4])
        self.assertEqual(c, ri.coverageCounts())

    def testOneIntervalEndingAtHitEndCoverageCounts(self):
        """
        If there is a single interval that ends at the end of the hit
        but doesn't start at zero, coverageCounts should return the correct
        result.
        """
        ri = ReadIntervals(10)
        ri.add(5, 10)
        c = Counter([5, 6, 7, 8, 9])
        self.assertEqual(c, ri.coverageCounts())

    def testOneIntervalInMiddleCoverageCounts(self):
        """
        If there is a single interval in the middle of the hit, coverageCounts
        should return the correct result.
        """
        ri = ReadIntervals(10)
        ri.add(5, 6)
        c = Counter([5])
        self.assertEqual(c, ri.coverageCounts())

    def testTwoOverlappingIntervalsInMiddleCoverageCounts(self):
        """
        If there are two overlapping intervals in the middle of the hit,
        coverageCounts should return the correct result.
        """
        ri = ReadIntervals(10)
        ri.add(5, 7)
        ri.add(6, 8)
        c = Counter([5, 6, 6, 7])
        self.assertEqual(c, ri.coverageCounts())

    # The following tests have the same setup as above, but they test the
    # coverage() method.

    def testEmptyCoverageOnZeroLengthSequence(self):
        """
        When no intervals are added, coverage should return 0.0
        """
        ri = ReadIntervals(0)
        self.assertEqual(0.0, ri.coverage())

    def testEmptyCoverage(self):
        """
        When no intervals are added, coverage should return 0.0
        """
        ri = ReadIntervals(100)
        self.assertEqual(0.0, ri.coverage())

    def testOneIntervalExactCoveringCoverage(self):
        """
        If there is a single interval that spans the whole hit exactly,
        coverage should return 1.0.
        """
        ri = ReadIntervals(100)
        ri.add(0, 100)
        self.assertEqual(1.0, ri.coverage())

    def testOneIntervalCoveringAllExtendingLeftCoverage(self):
        """
        If there is a single interval that spans the whole hit, including
        going negative to the left, coverage should return 1.0.
        """
        ri = ReadIntervals(100)
        ri.add(-10, 100)
        self.assertEqual(1.0, ri.coverage())

    def testOneIntervalCoveringAllExtendingRightCoverage(self):
        """
        If there is a single interval that spans the whole hit, including
        going beyond the hit to the right, coverage should return 1.0.
        """
        ri = ReadIntervals(100)
        ri.add(0, 110)
        self.assertEqual(1.0, ri.coverage())

    def testOneIntervalCoveringAllExtendingBothCoverage(self):
        """
        If there is a single interval that spans the whole hit, including
        starting before zero and also going beyond the hit to the right,
        coverage should return 1.0
        """
        ri = ReadIntervals(100)
        ri.add(-10, 110)
        self.assertEqual(1.0, ri.coverage())

    def testOneIntervalStartingAtZeroCoverage(self):
        """
        If there is a single interval that starts at zero but doesn't
        cover the whole hit, coverage should return the correct value.
        """
        ri = ReadIntervals(100)
        ri.add(0, 50)
        self.assertEqual(0.5, ri.coverage())

    def testOneIntervalStartingBeforeZeroCoverage(self):
        """
        If there is a single interval that starts before zero but doesn't
        cover the whole hit, coverage should return the correct value.
        """
        ri = ReadIntervals(100)
        ri.add(-50, 50)
        self.assertEqual(0.5, ri.coverage())

    def testOneIntervalEndingAtHitEndCoverage(self):
        """
        If there is a single interval that ends at the end of the hit
        but doesn't start at zero, coverage should return the correct value.
        """
        ri = ReadIntervals(100)
        ri.add(50, 100)
        self.assertEqual(0.5, ri.coverage())

    def testOneIntervalEndingAfterHitEndCoverage(self):
        """
        If there is a single interval that ends after the end of the hit
        but doesn't start at zero, coverage should return the correct value.
        """
        ri = ReadIntervals(100)
        ri.add(50, 150)
        self.assertEqual(0.5, ri.coverage())

    def testOneIntervalInMiddleCoverage(self):
        """
        If there is a single interval in the middle of the hit, coverage
        should return the correct value.

        """
        ri = ReadIntervals(100)
        ri.add(50, 60)
        self.assertEqual(0.1, ri.coverage())

    def testTwoOverlappingIntervalsInMiddleCoverage(self):
        """
        If there are two overlapping intervals in the middle of the hit,
        coverage should return the correct value.
        """
        ri = ReadIntervals(100)
        ri.add(50, 60)
        ri.add(55, 70)
        self.assertEqual(0.2, ri.coverage())

    def testThreeOverlappingIntervalsInMiddleCoverage(self):
        """
        If there are three overlapping intervals in the middle of the hit,
        coverage should return the correct value.
        """
        ri = ReadIntervals(100)
        ri.add(50, 60)
        ri.add(65, 75)
        ri.add(55, 70)
        self.assertEqual(0.25, ri.coverage())

    def testPairOfTwoOverlappingIntervalsCoverage(self):
        """
        If there are two sets of two overlapping intervals in the middle of
        the hit, coverage should return the correct value.
        """
        ri = ReadIntervals(100)
        # First overlapping pair, 50-70.
        ri.add(50, 60)
        ri.add(55, 70)
        # First overlapping pair, 80-95.
        ri.add(80, 90)
        ri.add(85, 95)
        self.assertEqual(0.35, ri.coverage())

    def testOverlappingIntervalsThatCoverEverythingCoverage(self):
        """
        If there are sets of overlapping intervals that cover the whole hit,
        coverage should return the correct value.
        """
        ri = ReadIntervals(100)
        ri.add(-10, 20)
        ri.add(15, 40)
        ri.add(40, 70)
        ri.add(66, 89)
        ri.add(77, 93)
        ri.add(70, 110)
        self.assertEqual(1.0, ri.coverage())


class TestOffsetAdjuster(TestCase):
    """
    Tests for the OffsetAdjuster class.
    """

    def testEmpty(self):
        """
        When no intervals are given, the reduction should be for the full hit
        width.
        """
        adjuster = OffsetAdjuster()
        self.assertEqual([], adjuster.adjustments())
        self.assertEqual(0, adjuster.adjustOffset(0))

    def testNoReads(self):
        """
        When no reads are added to an interval, the reduction should be for
        the full hit width.
        """
        ri = ReadIntervals(64)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
                (64, 58)
            ],
            adjuster.adjustments())
        self.assertEqual(6, adjuster.adjustOffset(64))

    def testOneReadThatExactlyCoversHit(self):
        """
        When one read is given that exactly covers the hit, there should
        be no length reductions.
        """
        ri = ReadIntervals(106)
        ri.add(0, 106)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
            ],
            adjuster.adjustments())
        self.assertEqual(106, adjuster.adjustOffset(106))

    def testOneReadThatExceedsHitOnBothEnds(self):
        """
        When one read is given that exceeds the hit at both ends, there should
        be no length reductions.
        """
        ri = ReadIntervals(106)
        ri.add(-100, 200)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
            ],
            adjuster.adjustments())
        self.assertEqual(106, adjuster.adjustOffset(106))

    def testOneReadAtStart(self):
        """
        When one read is added to the start of an interval, there should be one
        reduction for the empty section after the read.
        """
        ri = ReadIntervals(228)
        ri.add(0, 100)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
                (228, 121),
            ],
            adjuster.adjustments())
        self.assertEqual(107, adjuster.adjustOffset(228))

    def testOneReadBeforeStart(self):
        """
        When one read is added to the start of an interval before zero, there
        should be one reduction for the empty section after the read.
        """
        ri = ReadIntervals(228)
        ri.add(-10, 100)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
                (228, 121),
            ],
            adjuster.adjustments())
        self.assertEqual(107, adjuster.adjustOffset(228))

    def testOneReadAtEnd(self):
        """
        When one read is added to the end of an interval, there should be one
        reduction for the empty section before the read.
        """
        ri = ReadIntervals(228)
        ri.add(128, 228)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
                (128, 121),
            ],
            adjuster.adjustments())
        self.assertEqual(107, adjuster.adjustOffset(228))

    def testOneReadAfterEnd(self):
        """
        When one read is added to the end of an interval, going beyond the end
        of the hit, there should be one reduction for the empty section before
        the read.
        """
        ri = ReadIntervals(228)
        ri.add(128, 250)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
                (128, 121),
            ],
            adjuster.adjustments())
        self.assertEqual(107, adjuster.adjustOffset(228))

    def testOneReadInMiddle(self):
        """
        When one read is added to the middle of an interval, there should be
        two reductions.
        """
        ri = ReadIntervals(106)
        ri.add(32, 42)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
                (32, 27),
                (106, 58),
            ],
            adjuster.adjustments())
        self.assertEqual(106 - 27 - 58, adjuster.adjustOffset(106))

    def testTwoReadsInMiddle(self):
        """
        When two reads are added to the middle of an interval, there should be
        three reductions (after first empty area, after 2nd empty area, after
        final empty area.
        """
        ri = ReadIntervals(132)
        ri.add(32, 42)
        ri.add(58, 68)
        adjuster = OffsetAdjuster(ri)
        self.assertEqual(
            [
                (32, 27),
                (58, 12),
                (132, 58),
            ],
            adjuster.adjustments())
        self.assertEqual(132 - 27 - 12 - 58, adjuster.adjustOffset(132))

        # Test an HSP at the beginning is unchanged.
        hsp = HSP(10, readEndInSubject=10, readStartInSubject=0,
                  subjectEnd=10, subjectStart=0)
        adjuster.adjustHSP(hsp)
        self.assertEqual(10, hsp.readEndInSubject)
        self.assertEqual(0, hsp.readStartInSubject)
        self.assertEqual(10, hsp.subjectEnd)
        self.assertEqual(0, hsp.subjectStart)

        # Test an HSP in the first read region.
        hsp = HSP(10, readEndInSubject=42, readStartInSubject=32,
                  subjectEnd=40, subjectStart=35)
        adjuster.adjustHSP(hsp)
        self.assertEqual(15, hsp.readEndInSubject)
        self.assertEqual(5, hsp.readStartInSubject)
        self.assertEqual(13, hsp.subjectEnd)
        self.assertEqual(8, hsp.subjectStart)

        # Test an HSP in the second read region.
        hsp = HSP(10, readEndInSubject=68, readStartInSubject=58,
                  subjectEnd=66, subjectStart=60)
        adjuster.adjustHSP(hsp)
        self.assertEqual(29, hsp.readEndInSubject)
        self.assertEqual(19, hsp.readStartInSubject)
        self.assertEqual(27, hsp.subjectEnd)
        self.assertEqual(21, hsp.subjectStart)
